"""
This is the interface to the backend.
It is used to make the backend easily interchangable.
All Functions that interface directly with the backend are prefixed with be_
Functions only used internal in this module will be prefixed _bif_
"""

import logging

import time
import os

from PyQt4 import QtGui, QtCore
import numpy as np

from backend import board
from backend import io
from backend import dataset
from groupedelements import Checkboxes, Buttons, Elements, live_plot_windows
import storage
import config
import kcgwidget as kcgw
from log import log

tr = kcgw.tr

livePlotData = None

class DummyStorage():
    def __init__(self):
        self.old_getattr = self.__getattr__
    def __getattr__(self, item):
        try:
            return self.old_getattr(item)
        except:
            print item+" was not set, None is returned"
            return None

def initStatus(st): # TODO: Better solution? or implement all necesseary status things
    st.continuous_read = False
    st.calibrated = False
    st.synced = False
    st.defaults_set = False
    st.status_text = tr("sw", "Ready")
    st.time_scan = False
    st.last_file = None
    st.board_connected = True
board.status = DummyStorage()
initStatus(board.status)
storage.tempstorage = DummyStorage()

# -----------[ register observer for header writing ] ----- # TODO: move this to the appropriate place
def update_header(status):
    _bif_status_readout()
    if not board.status.board_connected:
        return
    try:
        control = board.read_pci(1, '0x9040')[0]
        control_bits = '{0:032b}'.format(int(control, 16))
        if status:
            control_bits = control_bits[:3] + '1' + control_bits[4:]
        else:
            control_bits = control_bits[:3] + '0' + control_bits[4:]
        dec_val_bits = int(control_bits, 2)
        board.write_pci(hex(dec_val_bits), '0x9040')
    except board.BoardError as e:
        reason = str(e) if str(e) != '' else "Unknown"
        mb = QtGui.QMessageBox(QtGui.QMessageBox.Critical, "Board communication",
                                   "Was unable to write value to board!\nReason: " + reason)
        mb.exec_()
        del mb
        _bif_status_readout()
board.config.observe(None, update_header, 'header')


# -----------[ Backend Interface ]----------------------


def _bif_enable_wait_cursor():
    """
    Show the "Wait Cursor"
    """
    QtGui.QApplication.setOverrideCursor(QtGui.QCursor(QtCore.Qt.WaitCursor))
def _bif_disable_wait_cursor():
    """
    Show the normal Cursor
    """
    QtGui.QApplication.restoreOverrideCursor()

def _bif_continuous_read_is_enabled(popup_title_text=None):
    """
    Checks if continuous read is enabled and if yes shows a popup dialog to ask if it shall be disabled
    :popup_title_text: Text to display in the popup that asks to disable continuous read
    :return: bool (True if continuous read is enabled and not disabled in popup else False)
    """
    if board.status.continuous_read:
        popup = PopupDialog(tr("Dialog", "Continuous read is currently active!\nStop continuous read and proceed?"),
                            title=popup_title_text)
        popup.exec_()
        popup.deleteLater()
        if popup.get_return_value() is False:
            return False
        else:
            _bif_set_continuous_read_inactive()
            Checkboxes.setChecked("continuousread", False)
            return True
    else:
        return False

def _bif_status_readout():
    # TODO: NOTE: Problem with this is that certain buttons will be greyed out until other buttons are pressed
    # TODO: even if only the gui is restarted and the board is in the same state
    if kcgw.testing:
        return # TODO: IMPORTANT REMOVE THIS !!!!!
    if board.is_active():
        Buttons.setEnabled("after_start", True)
        Buttons.setEnabled("start_board", False)
        Buttons.setEnabled("continuous_read", True)

        # MenuItems.setEnabled("continuous_read", True)
    else:
        board.status.calibrated = False
        board.status.synced = False
        board.status.defaults_set = False
        Buttons.setEnabled("after_start", False)
        Buttons.setEnabled("continuous_read", False)
        # MenuItems.setEnabled("continuous_read", False)
    Buttons.setEnabled("synchronize", board.status.calibrated)
    Buttons.setEnabled("set_defaults", board.status.synced)
    Buttons.setEnabled("acquire", board.status.synced)
    Buttons.setEnabled("acquireTrigger", board.status.synced)
    # Elements.setEnabled("timing", board.status.defaults_set)
    Elements.setEnabled("timing", board.status.synced)

    registers = board.read_pci(3, '0x9050', decimal=True)
    storage.update_LED(registers)

backup_readout = _bif_status_readout

class PopupDialog(QtGui.QDialog):
    def __init__(self, text, title=None, parent=None):
        QtGui.QDialog.__init__(self, parent)
        self.text = text
        self.setWindowTitle(title if title else tr("Dialog", "User action required"))
        self.do_layout()
        self.return_value = False

    def do_layout(self):
        size = QtCore.QSize(200, 200)
        #self.setMaximumSize(size)
        self.setMinimumSize(size)
        box = QtGui.QVBoxLayout()
        self.text_label = QtGui.QLabel(self.text)
        self.text_label.setAlignment(QtCore.Qt.AlignCenter)
        box.addWidget(self.text_label)
        self.okay_btn = QtGui.QPushButton(tr("Button", "Ok"))
        self.okay_btn.setStyleSheet("padding: 15px;")
        self.okay_btn.clicked.connect(self.on_okay)
        box.addWidget(self.okay_btn)
        box.addSpacerItem(QtGui.QSpacerItem(1, 20))
        self.cancel_btn = QtGui.QPushButton(tr("Button", "Cancel"))
        self.cancel_btn.setStyleSheet("padding: 15px;")
        self.cancel_btn.clicked.connect(self.on_cancel)
        box.addWidget(self.cancel_btn)
        self.setLayout(box)

    def on_okay(self):
        self.return_value = True
        self.close()

    def on_cancel(self):
        self.close()

    def get_return_value(self):
        return self.return_value


def bk_start_board():
    _bif_enable_wait_cursor()
    log(additional="Starting Board - following values are probably default values")

    if _bif_continuous_read_is_enabled(tr("Button", "Start Board")):
        _bif_disable_wait_cursor()
        return

    try:
        if board.is_active():
            _bif_disable_wait_cursor()
            _bif_status_readout()
            return

        logging.info("##### Activating Board #####")
        board.start_dma()
        board.write_pci('0x20001000', '0x9100')
        board.write_pci('0x05', '0x9040')
        time.sleep(1.0)
        dialog1 = PopupDialog(tr("Button", "Switch On the power supply --> FIRST <--"))
        dialog1.exec_()
        dialog1.deleteLater()

        if not dialog1.get_return_value():
            logging.error("Starting procedure canceled")
            _bif_disable_wait_cursor()
            return

        logging.info("Switch ON T/Hs")
        board.write_pci('0x3C1', '0x9040')
        logging.info("9040:  " + str(board.read_pci(1, '0x9040')[0]))
        time.sleep(0.1)
        dialog2 = PopupDialog(tr("Dialog", "Switch On the power supply --> SECOND <--"))
        dialog2.exec_()
        dialog2.deleteLater()

        if not dialog2.get_return_value():
            logging.info("Starting procedure canceled")
            _bif_disable_wait_cursor()
            return

        logging.info("Switch ON ADCs")
        board.write_pci('0x3F1', '0x9040')
        logging.info("9040:  " + str(board.read_pci(1, '0x9040')[0]))
        time.sleep(0.1)

        board.write_pci('0x3F0', '0x9040')
        logging.info("9040:  " + str(board.read_pci(1, '0x9040')[0]))
        time.sleep(1.0)

        logging.info("Board started successfully!")
    except board.BoardError as e:
        logging.error("Starting board failed: {}".format(str(e)))


    board.config.update('header', board.config.get('header'))
    _bif_disable_wait_cursor()
    _bif_status_readout()

class _bif_ProgressBar(QtGui.QProgressBar):
    def __init__(self, min, max, text):
        super(_bif_ProgressBar, self).__init__()
        self.setRange(min, max)
        self.setMaximumHeight(18)
        kcgw.statusbar.clearMessage()
        self.label = QtGui.QLabel(text)
        kcgw.statusbar.insertWidget(0, self.label)
        kcgw.statusbar.insertWidget(1, self)
        self.setValue(0)
        QtGui.qApp.processEvents()
    def remove(self, timeout=None):
        def remove_progressbar():
            kcgw.statusbar.removeWidget(self)
            kcgw.statusbar.removeWidget(self.label)
            kcgw.statusbar.showMessage(board.status.status_text)
            self.destroy()
        if timeout:
            QtCore.QTimer.singleShot(timeout, remove_progressbar)
        else:
            remove_progressbar()

thread = None
cal = None
def bk_calibrate(do_the_rest=None):
    log(additional="Calibrate")
    global thread
    global cal
    if thread is not None:
        logging.info("Calibration already running")
        return
    _bif_enable_wait_cursor()
    if _bif_continuous_read_is_enabled(tr("Button", "Calibrate Board")):
        _bif_disable_wait_cursor()
        return
    progressbar = _bif_ProgressBar(0, 15, tr("sw", "Calibrating"))
    class Calibrate(QtCore.QObject):
        update_progressbar_signal = QtCore.pyqtSignal(int)
        finished = QtCore.pyqtSignal()
        def calibrate(self):
            try:
                logging.info("##### Calibrating Board #####")
                logging.info("Removing Board Reset")
                board.write_pci('0x2000003f0', '0x9040')
                time.sleep(0.5)
                # progressbar.setValue(1)
                self.update_progressbar_signal.emit(1)

                logging.info("SPI Fanout Programming...")
                board.write_pci('0x083', '0x9068')
                time.sleep(0.5)
                # progressbar.setValue(2)
                self.update_progressbar_signal.emit(2)
                board.write_pci('0x6a2', '0x9068')
                time.sleep(0.5)
                # progressbar.setValue(3)
                self.update_progressbar_signal.emit(3)

                logging.info("PLL calibration start...")
                logging.info("PLL Reset...")
                board.write_pci('0x80000000', '0x9060')
                time.sleep(0.5)
                # progressbar.setValue(4)
                self.update_progressbar_signal.emit(4)
                logging.info("Set CH_0 (FPGA) clock FPGA ...")
                board.write_pci('0x00050000', '0x9060')
                time.sleep(0.5)
                # progressbar.setValue(5)
                self.update_progressbar_signal.emit(5)

                logging.info("Set CH_3 clock fanout ...")
                board.write_pci('0x00050003', '0x9060')
                time.sleep(0.5)
                # progressbar.setValue(6)
                self.update_progressbar_signal.emit(6)

                logging.info("Set CH_4 clock ADC 1 ...")
                board.write_pci('0x00050004', '0x9060')
                time.sleep(0.5)
                # progressbar.setValue(7)
                self.update_progressbar_signal.emit(7)

                logging.info("Set CH_5 clock ADC 2 ...")
                board.write_pci('0x00050005', '0x9060')
                time.sleep(0.5)
                # progressbar.setValue(8)
                self.update_progressbar_signal.emit(8)

                logging.info("Set CH_6 clock ADC 3 ...")
                board.write_pci('0x00050006', '0x9060')
                time.sleep(0.5)
                # progressbar.setValue(9)
                self.update_progressbar_signal.emit(9)

                logging.info("Set CH_7 clock ADC 4 ...")
                board.write_pci('0x00050007', '0x9060')
                time.sleep(0.5)
                # progressbar.setValue(10)
                self.update_progressbar_signal.emit(10)

                logging.info("Set R8 ...")
                board.write_pci('0x10000908', '0x9060')
                time.sleep(0.5)
                # progressbar.setValue(11)
                self.update_progressbar_signal.emit(11)

                logging.info("Set R11 ...")
                board.write_pci('0x0082800B', '0x9060')
                time.sleep(0.5)
                # progressbar.setValue(12)
                self.update_progressbar_signal.emit(12)

                logging.info("Set R13 ...")
                board.write_pci('0x029F400D', '0x9060')
                time.sleep(0.5)
                # progressbar.setValue(13)
                self.update_progressbar_signal.emit(13)

                logging.info("Set R14 (F_out and Global_EN => ON) ...")
                board.write_pci('0x0830040E', '0x9060')
                time.sleep(0.5)
                # progressbar.setValue(14)
                self.update_progressbar_signal.emit(14)

                logging.info("Set R15 ...")
                board.write_pci('0xD000100F', '0x9060')
                time.sleep(0.5)
                # progressbar.setValue(15)
                self.update_progressbar_signal.emit(15)

                board.status.calibrated = True
            except board.BoardError as e:
                logging.error("Calibration failed: {}".format(str(e)))
                _bif_disable_wait_cursor()
                # self.do_status_readout()
                _bif_status_readout()
                self.finished.emit()
                return

            logging.info("Board Calibration successful!")
            _bif_status_readout()
            self.finished.emit()

    def thread_quit():
        global thread
        progressbar.remove(0)
        # thread.quit()
        # thread.wait()
        thread = None
        _bif_disable_wait_cursor()
        if do_the_rest: # execute sync and set defaults (this is set if prepare board was pressed)
            do_the_rest()
    # thread = QtCore.QThread()
    cal = Calibrate()
    cal.update_progressbar_signal.connect(progressbar.setValue)
    cal.finished.connect(thread_quit)
    # cal.moveToThread(thread)
    # thread.started.connect(cal.calibrate)
    # thread.start()
    cal.calibrate()

def bk_sync_board():
    _bif_enable_wait_cursor()
    log(additional="Synchronize")
    if _bif_continuous_read_is_enabled(tr("Dialog", "Synchronize Board")):
        _bif_disable_wait_cursor()
        return

    try:
        progressbar = _bif_ProgressBar(0, 100, tr("sw", "Synchronizing"))
        logging.info("##### Synchronize PLLs #####")
        logging.info("Send the PLL sync signals ...")
        board.write_pci('0x1003f0', '0x9040')
        # time.sleep(1.0)
        for i in range(1, 101):
            time.sleep(0.01)
            progressbar.setValue(i)
        logging.info("Done!")
        board.write_pci('0x0003f0', '0x9040')
        progressbar.remove(0)
    except board.BoardError as e:
        logging.error("Synchronization failed: {}".format(str(e)))
        progressbar.remove(0)
        _bif_disable_wait_cursor()
        _bif_status_readout()
        return

    logging.info("Board synchronization successful!")
    # self.set_defaults_button.setEnabled(True)
    board.status.synced = True
    _bif_disable_wait_cursor()
    _bif_status_readout()

def bk_set_defaults():
    _bif_enable_wait_cursor()
    log(additional="Set Default Values")
    if _bif_continuous_read_is_enabled(tr("Dialog", "Synchronize Board")):
        _bif_disable_wait_cursor()
        return

    try:
        progressbar = _bif_ProgressBar(0, 12, tr("sw", "Setting Defaults"))
        logging.info("##### Setting default Values #####")
        logging.info("Set Defaults delay value in the board...")
        #Set_FPGA_clock_delay.sh 0
        board.config.set_fpga_delay(board.config.get('fpga_delay'))
        time.sleep(0.1)
        progressbar.setValue(1)
        #Set_Delay_chip.sh 16 16 16 16
        factors = [board.config.get('chip_1_delay'), board.config.get('chip_2_delay'),
                   board.config.get('chip_3_delay'), board.config.get('chip_4_delay')]
        board.config.set_chip_delay([0, 1, 2, 3], factors)
        # ---[ ugly ]-----
        time.sleep(0.1)
        progressbar.setValue(2)
        time.sleep(0.1)
        progressbar.setValue(3)
        time.sleep(0.1)
        progressbar.setValue(4)
        time.sleep(0.1)
        progressbar.setValue(5)
        time.sleep(0.1)
        progressbar.setValue(6)
        time.sleep(0.1)
        progressbar.setValue(7)
        #-----------------
        #Set_TH_Delay.sh 12
        board.config.set_th_delay(board.config.get('th_delay'))
        time.sleep(0.1)
        progressbar.setValue(8)
        #Set_ADC_1_Delay.sh 4
        board.config.set_adc_delay(0, board.config.get('adc_1_delay'))
        time.sleep(0.1)
        progressbar.setValue(9)
        #Set_ADC_2_Delay.sh 4
        board.config.set_adc_delay(1, board.config.get('adc_2_delay'))
        time.sleep(0.1)
        progressbar.setValue(10)
        #Set_ADC_3_Delay.sh 4
        board.config.set_adc_delay(2, board.config.get('adc_3_delay'))
        time.sleep(0.1)
        progressbar.setValue(11)
        #Set_ADC_4_Delay.sh 4
        board.config.set_adc_delay(3, board.config.get('adc_4_delay'))
        time.sleep(0.1)
        progressbar.setValue(12)

        board.write_pci('{0:08x}'.format(board.config.get('orbits_observe')), '0x9020')
        # acqs.setValueSilent('orbits_observe', board.config.get('orbits_observe'))


        board.write_pci('{0:08x}'.format(board.config.get('orbits_skip')), '0x9028')
        # acqs.setValueSilent('orbits_skip', board.config.get('orbits_skip'))
        board.status.defaults_set = True
        progressbar.remove(0)

    except board.BoardError as e:
        logging.error("Setting defaults failed: {}".format(str(e)))
        progressbar.remove(0)
        _bif_disable_wait_cursor()
        _bif_status_readout()
        return

    logging.info("Default values set successfully!")
    # Elements.setEnabled('timing', True)
    _bif_disable_wait_cursor()
    _bif_status_readout()


def bk_stop_board():
    _bif_enable_wait_cursor()
    log(additional="Stop Board")
    # if not self._check_for_no_continuous_read():
    #     _bif_disable_wait_cursor()
    #     return
    if _bif_continuous_read_is_enabled(tr("Dialog", "Soft Reset Board")):
        _bif_disable_wait_cursor()
        return

    try:
        logging.info("##### Switching Off Board #####")
        board.write_pci('0x01', '0x9040')
        board.stop_dma()
        cmd = ['pci', '--list-dma-engines']
        logging.info(board.safe_call(cmd))
        time.sleep(0.5)
    except board.BoardError as e:
        logging.error("Sequence failed: {}".format(str(e)))
        _bif_disable_wait_cursor()
        _bif_status_readout()
        return

    logging.info("Board switched off successfully!")
    # self.start_board_button.setEnabled(True)
    # self.calibrate_board_button.setEnabled(False)
    # self.sync_board_button.setEnabled(False)
    # self.set_defaults_button.setEnabled(False)
    # self.set_adc_delay_spinboxes_active(False)
    Buttons.setEnabled("after_start", False)
    Buttons.setEnabled("start_board", True)
    board.status.calibrated = False
    _bif_disable_wait_cursor()
    _bif_status_readout()

def bk_soft_reset():
    _bif_enable_wait_cursor()
    log(additional="Soft Reset")
    # if not self._check_for_no_continuous_read():
    #     _bif_disable_wait_cursor()
    #     return
    if _bif_continuous_read_is_enabled(tr("Dialog", "Soft Reset Board")):
        _bif_disable_wait_cursor()
        return

    try:
        logging.info("Soft-Resetting Board ...")
        board.write_pci('0x1', '0x9040', hex_mask='0x1')
        time.sleep(1)
        board.write_pci('0x0', '0x9040', hex_mask='0x1')
    except board.BoardError as e:
        logging.error("Sequence failed: {}".format(str(e)))
        _bif_disable_wait_cursor()
        _bif_status_readout()
        return

    logging.info("Soft-Reset successful.")
    _bif_disable_wait_cursor()
    _bif_status_readout()
    update_header(True) # Reset Header if Header bit is erased after soft reset

def bk_update_config(key, value):
    board.config.update(key, value)
def bk_get_config(key):
    return board.config.get(key)
def bk_change_num_of_orbits(value, silent=False):
    """
    Send new number of orbits to board and update in config
    :param value: the value to send
    :param silent: update config silent (that is: do not notify observers)
    :return: -
    """
    if silent:
        board.config.updateSilent("orbits_observe", value)
    else:
        board.config.update("orbits_observe", value)
    hex_str = hex(value)
    try:
        board.write_pci(hex_str, '0x9020')
    except board.BoardError as e:
        logging.error("Setting orbits failed: {}".format(str(e)))
        _bif_status_readout()
def bk_change_num_of_skipped_orbits(value, silent=False):
    """
    Send new number of orbits to skip to board and update in config
    :param value: the value to send
    :param silent: update config silent (that is: do not notify observers)
    :return: -
    """
    if silent:
        board.config.updateSilent("orbits_skip", value)
    else:
        board.config.update("orbits_skip", value)
    hex_str = hex(value)
    try:
        board.write_pci(hex_str, '0x9028')
    except board.BoardError as e:
        logging.error("Setting skip orbits failed: {}".format(str(e)))
        _bif_status_readout()
def bk_change_count(value, silent=False):
    if silent:
        board.config.updateSilent("orbits_count", value)
    else:
        board.config.update("orbits_count", value)
def bk_change_wait(value, silent=False):
    if silent:
        board.config.updateSilent("orbits_wait_time", value)
    else:
        board.config.update("orbits_wait_time", value)
def bk_change_build_spectrograms(value, silent=False):
    if silent:
        board.config.updateSilent("build_spectrograms", value)
    else:
        board.config.update("build_spectrograms", value)
def bk_change_pilot_bunch(value, silent=False):
    if silent:
        board.config.updateSilent("pilot_bunch", value)
    else:
        board.config.update("pilot_bunch", value)


def _bif_iterate_spectrograms(path):
    if not os.path.isdir(str(path)):
        return

    transform = dataset.data.fft(1, frm=0, to=-1)
    for i in range(config.bunches_per_turn-1):
        filename = os.path.join(storage.storage.save_location, storage.storage.subdirname, str(path), "%i.hsp" % i)
        write_header = False
        if not os.path.isfile(filename):
            write_header = True
        f = open(filename, 'ab')
        if write_header:
            f.write("#hsp\n")  # heb spectrogram magic number
            #f.write(b'\x40')  # 64bit
            #f.write('64')  # 64bit
            #f.write(sytruct.pack('>I', self.number_of_skipped_orbits_spinbox.value()))
            f.write("#"+str(board.config.get("orbits_skip")))
            f.write("\n")
        line = transform[i,:]
        # f.write(struct.pack('>I', len(line)))
        #f.write(str(len(line)))
        f.write('{:0.3f} '.format(time.time()))
        #data = pArr('d', line)
        #data.tofile(f)
        for e in line:
            f.write("%s "%np.absolute(e))
        f.write("\n")
        f.close()

def _bif_read_data_and_save():
    now = time.time()
    if not os.path.isdir(str(storage.storage.save_location + '/' + storage.storage.subdirname)):
        os.makedirs(str(storage.storage.save_location + '/' + storage.storage.subdirname))
    filename = storage.storage.save_location + '/' + storage.storage.subdirname+'/{:0.3f}.out'.format(now)
    board.status.last_file = filename

    try:
        simulate = board.config.get("pilot_bunch")
        board.acquire_data(filename, simulate=simulate, duration=2.0)
        _bif_read_and_update_data_from_file(filename)
    except board.BoardError as e:
        logging.error("Reading failed: {}".format(str(e)))

def _bif_read_and_update_data_from_file(filename):
    _bif_read_and_update(io.read_from_file, str(filename))

def _bif_read_and_update_data_from_string(raw_data):
    _bif_read_and_update(io.read_from_string, raw_data)

def _bif_read_and_update(read_func, *args):
    _bif_enable_wait_cursor()

    header = board.config.get('header')
    # TODO: force_read: meaning ignore cache and read new -> in the old gui this was a commandline option
    # TODO: cache_data: meaning cache already processed numpy data -> in the old gui this was a commandline option
    if live_plot_windows.hasWindows():
        data = read_func(*args, force=False, header=header, cache=False)

        if not io.is_data_consistent(data):
            kcgw.consistency_updater(False)
            # self.data_consistency_led.set_color((0, 190, 0))
            kcgw.statusbar.showMessage(tr("Dialog", "Data is inconsistent!"))
            if read_func == io.read_from_string:
                logging.info("Data is inconsistent")
            else:
                logging.info("Data is inconsistent - file: " + args[0])
        else:
            kcgw.consistency_updater(True)

        for plotwin in live_plot_windows.getWindows():
            plotwin.plot_live(data=data)
    else:
        kcgw.consistency_updater(None)

    _bif_disable_wait_cursor()


def bk_acquire():
    if board.status.acquisition == True:
        log(additional="Manually Stopped Acquisition\nPerformed Acquisitions: " + str(storage.tempstorage.current_acquisition))
        _bif_stop_acquisition()
    else:
        _bif_start_acquisition()


def _bif_stop_acquisition():
    board.status.acquisition = False
    board.acquisition_progressbar.remove(0)
    board.acquisition_timer.stop()
    for elem in Elements.getElements("acquireTrigger"):
        elem.setIcon(QtGui.QIcon(config.startIcon))
        elem.setText(tr("Button", "Start Acquisition"))
    # self.continuous_read_checkbox.setEnabled(True)
    # self.acquisition_progressbar.reset()
    Elements.setEnabled('acquire', True)


def _bif_start_acquisition():
    log(additional="Started Acquisition")
    board.status.acquisition = True
    Elements.setEnabled('acquire', False)
    for elem in Elements.getElements("acquireTrigger"):
        elem.setIcon(QtGui.QIcon(config.stopIcon))
        elem.setText(tr("Button", "Stop Acquisition"))
    board.acquisition_timer = QtCore.QTimer()
    num_acquisitions = board.config.get("orbits_count")
    acquisition_progressbar = _bif_ProgressBar(0, num_acquisitions, tr("sw", "Acquiring"))
    board.acquisition_progressbar = acquisition_progressbar


    # We increase already once because we do a single acquisition before the
    # timer is started, otherwise we have to wait until the timer fires the
    # first time.
    storage.tempstorage.current_acquisition = 1
    _bif_read_data_and_save()

    if board.config.get("build_spectrograms"):
        spectrogram_dir = storage.storage.save_location + '/' + storage.storage.subdirname+"/spectrograms_{:0.3f}".format(time.time())
        os.makedirs(spectrogram_dir)
        _bif_iterate_spectrograms(spectrogram_dir) # TODO: not here?

    acquisition_progressbar.setValue(storage.tempstorage.current_acquisition)

    def on_timeout():
        if storage.tempstorage.current_acquisition < num_acquisitions:
            storage.tempstorage.current_acquisition += 1
            acquisition_progressbar.setValue(storage.tempstorage.current_acquisition)
            _bif_read_data_and_save()
            if board.config.get("build_spectrograms"):
                _bif_iterate_spectrograms(spectrogram_dir) # TODO: not here ?
        else:
            log(additional="Stopped Acquisition")
            _bif_stop_acquisition()
            acquisition_progressbar.remove(0)

    board.acquisition_timer.timeout.connect(on_timeout)
    board.acquisition_timer.start(board.config.get('orbits_wait_time') * 1000)


def bk_single_read():
    Elements.setEnabled("acquire", False)
    _bif_read_data_and_save()
    log(additional="Single Read\nFilename: "+board.status.last_file.split('/')[-1])
    Elements.setEnabled("acquire", True)


def _bif_set_continuous_read_active():
    Elements.setEnabled("acquire", False)
    Elements.setEnabled("continuous_read", True)
    board.status.continuous_read = True


def _bif_set_continuous_read_inactive():
    if board.status.continuous_read:
        board.status.continuous_read = False
        board.continuous_read_timer.stop()
    Elements.setEnabled('acquire', True)


def bk_continuous_read(interval=100):
    if not board.status.continuous_read:
        _bif_set_continuous_read_active()
        _bif_continuous_read(interval)
    else:
        _bif_set_continuous_read_inactive()


def _bif_read_data():
    try:
        if board.config.get('pilot_bunch'):
            board.start_pilot_bunch_emulator()

        board.start_acquisition()
        board.wait_for_revolutions()
        board.stop_acquisition()
        board.enable_transfer()
        cmd = ['pci', '-r', 'dma0', '-o', '/dev/stdout', '--multipacket']
        data_raw = board.safe_call(cmd)
        board.flush_dma()
        # Sadly, the PCI software also outputs the Information, that it has
        # written data to the desired file.  This is recorded by our
        # software and needs to be removed, so we split the returned string
        # at the 'Writting data to file blah, blah" Btw: YES this
        # 'Writting' is actually correct. No, i don't know why it is
        # grammatically wrong in the first place :D
        data_split = data_raw.split('Writting')[0]
        _bif_read_and_update_data_from_string(data_split)
    except board.BoardError as e:
        logging.error("Reading failed: {}".format(str(e)))


def _bif_continuous_read(interval=100):
    board.continuous_read_timer = QtCore.QTimer()
    logging.info("Start continuous read")

    def continuous_read_step():
        if board.status.continuous_read:
            _bif_read_data()
            # board.continuous_read_timer.start(interval)
        else:
            board.continuous_read_timer.stop()

    board.continuous_read_timer.timeout.connect(continuous_read_step)
    board.continuous_read_timer.start(interval)


def bk_board_connected():
    return board.is_conneced()


def bk_get_temperature():
    fpga_temp_raw_hex = board.read_pci(1,'0x9110')[0]
    fpga_temp_raw_hex = fpga_temp_raw_hex[-3:]
    fpga_temp_raw_bin = '{0:012b}'.format(int(fpga_temp_raw_hex, 16))
    fpga_temp_encoded = board.get_dec_from_bits(fpga_temp_raw_bin, 9, 0)
    fpga_temp_celsius = '{0:2.2f}'.format(((fpga_temp_encoded*503.975)/1024)-273.15)
    return fpga_temp_celsius

backup_get_temp = bk_get_temperature

def bk_time_scan(c_frm, c_to, f_frm, f_to, ts_pbar, plot_func):
    if board.status.time_scan:
        _bif_stop_time_scan(ts_pbar)
    else:
        _bif_start_time_scan(c_frm, c_to, f_frm, f_to, ts_pbar, plot_func)


def _bif_stop_time_scan(ts_pbar):
    Elements.getElements("start_time_scan")[0].setText(tr("Button", "Start time scan"))
    board.status.time_scan = False
    board.config.set_delay(storage.th_old)
    board.config.set_chip_delay([0, 1, 2, 3], [storage.adc_1_old, storage.adc_2_old, storage.adc_3_old, storage.adc_4_old])
    # self.enable_acquisition_input()
    # self.continuous_read_checkbox.setEnabled(True)
    # self.scan_button.setEnabled(True)
    # self.scan_stop_button.setEnabled(False)
    # self.set_adc_delay_spinboxes_active(True)
    # TODO: activate/deactivate inputs
    ts_pbar.reset()

tst = None # Ugly but necessary for the thread not to be killed when function ends (which is almost immediately after start)
thread_ts = None
def _bif_start_time_scan(c_frm, c_to, f_frm, f_to, timescan_progressbar, plot_func):
    if thread_ts is not None:
        logging.info("Time scan already running")
        return
    board.status.time_scan = True
    Elements.getElements("start_time_scan")[0].setText(tr("Button", "Stop time scan"))
    # c_frm = self.time_scan_coarse_from.value()
    # c_to = self.time_scan_coarse_to.value()
    if c_frm > c_to:
        logging.info('Coarse Scan Interval is invalid: (%i > %i)' % (c_frm, c_to))
        return
    # f_frm = self.time_scan_fine_from.value()
    # f_to = self.time_scan_fine_to.value()
    if f_frm > f_to:
        logging.info('Fine Scan Interval is invalid: (%i > %i)' % (f_frm, f_to))
        return

    # self.disable_acquisition_input()
    # self.continuous_read_checkbox.setEnabled(False)
    # self.scan_button.setEnabled(False)
    # self.set_adc_delay_spinboxes_active(False)
    # TODO: set active/inactive etc

    # self.canvas.figure.clear()

    storage.th_old = board.config.get('th_delay')
    storage.adc_1_old = board.config.get('adc_1_delay')
    storage.adc_2_old = board.config.get('adc_2_delay')
    storage.adc_3_old = board.config.get('adc_3_delay')
    storage.adc_4_old = board.config.get('adc_4_delay')

    minimum = [None, None, None, None]
    maximum = np.zeros((4, 3))
    heatmap = np.zeros((4, (f_to - f_frm + 1), (c_to - c_frm + 1)))
    timescan_progressbar.setRange(1, ((f_to - f_frm) + 1) * ((c_to - c_frm) + 1))

    # self.scan_stop_button.clicked.connect(time_scan_stop)
    # self.scan_stop_button.setEnabled(True)

    class thread_time_scan(QtCore.QObject):
        pbarSignal = QtCore.pyqtSignal(int)
        stopSignal = QtCore.pyqtSignal()
        finished = QtCore.pyqtSignal()
        def __init__(self, c_frm, c_to, f_frm, f_to, timescan_progressbar):
            super(thread_time_scan, self).__init__()
            self.c_frm = c_frm
            self.c_to = c_to
            self.f_frm = f_frm
            self.f_to = f_to
            self.timescan_progressbar = timescan_progressbar
        def time_scan(self):
            c_step = 0
            for coarse in range(self.c_frm, self.c_to + 1):
                try:
                    board.config.set_delay(coarse)
                except board.BoardError as e:
                    # QtGui.QMessageBox.critical(self, "Board communication", str(e))
                    # _bif_stop_time_scan(self.timescan_progressbar)
                    self.stopSignal.emit()
                    self.finished.emit()
                    return

                f_step = 0
                for fine in range(self.f_frm, self.f_to + 1):
                    board.config.set_chip_delay([0, 1, 2, 3], [fine, fine, fine, fine])

                    try:
                        if bk_get_config('pilot_bunch') is True:
                            board.start_pilot_bunch_emulator()

                        board.start_acquisition()
                        board.wait_for_revolutions()
                        board.stop_acquisition()
                        board.enable_transfer()

                        # -----------[ IMPORTANT ]---------------------
                        if not kcgw.testing:
                            cmd = ['pci', '-r', 'dma0', '-o', '/dev/stdout', '--multipacket']
                            data_raw = board.safe_call(cmd)
                            board.flush_dma()
                        # ----------------------------------------------
                        else:
                            f_name = "/home/blaxxun/Documents/Hiwi/KaptureSimulator/timescan/" + str(coarse) + "_" + str(fine) + ".str"
                            f = open(f_name, 'r')
                            data_raw = f.read()

                        #Sadly, the PCI software also outputs the Information, that it has written data to the desired file.
                        #This is recorded by our software and needs to be removed, so we split the returned string at the
                        # "Writting data to file blah, blah" string.
                        #Btw: YES this 'Writting' is actually correct. Its a typo in the 'pci' driver...
                        data_split = data_raw.split('Writting')[0]
                        data = io.read_from_string(data_split, force=True, cache=False)
                    except board.BoardError as e:
                        QtGui.QMessageBox.critical(self, "Board communication", str(e)) # TODO: remove this if moved to thread again
                        # time_scan_stop()
                        # _bif_stop_time_scan(self.timescan_progressbar)
                        self.stopSignal.emit()
                        self.finished.emit()
                        return

                    for adc in range(4):
                        buckets = data.array[:, adc:adc+1].reshape(-1)
                        heatmap[adc, f_step, c_step] = float(buckets.sum())/buckets.shape[0]
                        if heatmap[adc, f_step, c_step] > maximum[adc, 0]:
                            maximum[adc, 0] = heatmap[adc, f_step, c_step]
                            maximum[adc, 1] = coarse
                            maximum[adc, 2] = fine
                        if minimum[adc] is None or minimum[adc] > heatmap[adc, f_step, c_step]:
                            minimum[adc] = heatmap[adc, f_step, c_step]

                    # timescan_progressbar.setValue(((c_step * (f_to - f_frm + 1)) + f_step) + 1)
                    self.pbarSignal.emit(((c_step * (f_to - f_frm + 1)) + f_step) + 1)

                    #GUI is blocked in our tight loop. Give it an opportunity to handle events
                    QtGui.QApplication.processEvents() # remove this if moved to thread again
                    if board.status.time_scan is False:
                        # Time Scan Stop is already performed by button press. Nothing else to do but leave
                        self.finished.emit()
                        return
                    f_step += 1
                c_step += 1
            self.finished.emit()

    def finished(timescan_progressbar):
        # global thread_ts
        # thread_ts.quit()
        # thread_ts.wait()
        # thread_ts = None
        _bif_stop_time_scan(timescan_progressbar)
        m = [np.min(heatmap[heatmap != 0]), np.max(heatmap)] # this gives the same levels for all 4 adcs
        plot_func(heatmap, levels=m, newTitle=str(tr("sw", "Coarserange:{c_f}-{c_t} ; Finerange:{f_f}-{f_t}")).format(
                                                                            c_f=c_frm,
                                                                            c_t=c_to,
                                                                            f_f=f_frm,
                                                                            f_t=f_to),
                  maxima = maximum
                  )

        now = time.time()
        if not os.path.isdir(str(storage.storage.save_location + '/' + storage.storage.subdirname + '/timescan')):
            os.makedirs((storage.storage.save_location + '/' + storage.storage.subdirname + '/timescan'))
        filename = storage.storage.save_location + '/' + storage.storage.subdirname+'/timescan/timescan_{:0.3f}.out'.format(now)
        # filename = 'timescan_{:0.3f}.out'.format(now)
        f = open(filename, 'wr')

        for adc in range(4):
            f.write("#ADC_%s\n" % adc)
            for coarse, curr_cor in enumerate(np.transpose(heatmap[adc])):
                for fine, value in enumerate(curr_cor):
                    f.write("%i;%i;%f\n" % ((coarse+c_frm), (fine+f_frm), value))
            f.write('\n')

        f.close()
        f = open(filename+'.gnuplot', 'wr')
        f.write('set datafile separator ";"\n')
        f.write('set multiplot layout 2,2\n')
        f.write('unset key\n')
        for i in range(4):
            f.write('set label 1 "ADC_%i" at graph 0.7,0.95 font ",8"\n'%(i+1))
            f.write('plot "%s" every :::%i::%i using 3 with lines\n'%(filename, i, i))
        f.write('unset multiplot\n')
        f.close()


        return

    global tst
    # global thread_ts
    # thread_ts = None
    # thread_ts = QtCore.QThread()
    tst = thread_time_scan(c_frm, c_to, f_frm, f_to, timescan_progressbar)
    tst.pbarSignal.connect(timescan_progressbar.setValue)
    tst.finished.connect(lambda: finished(timescan_progressbar))
    tst.stopSignal.connect(lambda: _bif_stop_time_scan(timescan_progressbar))
    # tst.moveToThread(thread_ts)
    # thread_ts.started.connect(tst.time_scan)
    # thread_ts.start()
    tst.time_scan()


def bk_check_for_board():
    """
    Check if board is connected
    Also overrides the _bif_status_readout function with a function that does nothing (suppresses read attempts that
    generate errors - if no board is connected, there is nothing to read from)
    Also overrides the bk_get_temperature function as of the same reasons
    :return: -
    """
    global _bif_status_readout, bk_get_temperature
    board_status = bk_board_connected()
    if board_status:
        if not board.status.board_connected:
            _bif_status_readout = backup_readout
            bk_get_temperature = backup_get_temp
        board.status.board_connected = True

    else:
        Elements.setEnabled('no_board', False)

        def do_nothing():
            pass

        def no_temp():
            return "-"
        _bif_status_readout = do_nothing
        bk_get_temperature = no_temp
        board.status.board_connected = False
        if board_status == False:
            board.status.status_text = tr("sw", "Board not connected")
        elif board_status == None:
            board.status.status_text = tr("sw", "Software Interface not found")
